#!/usr/bin/env python

from email import message_from_file
from email.mime.text import MIMEText
from smtplib import SMTP
from vCalendar import parse, iterwrite, ParseError, SECTION_TYPES
import imip_store
import sys

try:
    from cStringIO import StringIO
except ImportError:
    from StringIO import StringIO

# Postfix exit codes.

EX_USAGE        = 64
EX_DATAERR      = 65
EX_NOINPUT      = 66
EX_NOUSER       = 67
EX_NOHOST       = 68
EX_UNAVAILABLE  = 69
EX_SOFTWARE     = 70
EX_OSERR        = 71
EX_OSFILE       = 72
EX_CANTCREAT    = 73
EX_IOERR        = 74
EX_TEMPFAIL     = 75
EX_PROTOCOL     = 76
EX_NOPERM       = 77
EX_CONFIG       = 78

# Permitted iTIP content types.

itip_content_types = [
    "text/calendar",                        # from RFC 6047
    "text/x-vcalendar", "application/ics",  # other possibilities
    ]

# Sending of outgoing messages.

def sendmail(sender, recipients, data):
    smtp = SMTP("localhost")
    smtp.sendmail(sender, recipients, data)
    smtp.quit()

# Processing of incoming messages.

def process(f, args):

    "Process content from the stream 'f' accompanied by the given 'args'."

    recipient = args[0]

    msg = message_from_file(f)
    print >>open("/tmp/imip.txt", "a"), "----"
    print >>open("/tmp/imip.txt", "a"), args
    print >>open("/tmp/imip.txt", "a"), "----"
    print >>open("/tmp/imip.txt", "a"), msg.as_string()
    print >>open("/tmp/imip.txt", "a")

    # Handle messages with iTIP parts.

    for part in msg.walk():
        if part.get_content_type() in itip_content_types and \
            part.get_param("method"):

            handle_itip_part(part, recipient)

def get_itip_elements(elements):
    d = {}
    for name, attr, value in elements:
        if not d.has_key(name):
            d[name] = []
        if name in SECTION_TYPES:
            d[name].append((attr, get_itip_elements(value)))
        else:
            d[name].append((attr, value))
    return d

def get_attr_value(d, name, single=True):
    if d.has_key(name):
        values = d[name]
        if single and len(values) == 1:
            return values[0]
        else:
            return values
    else:
        return None

def get_value(d, name, single=True):
    if d.has_key(name):
        values = d[name]
        if single and len(values) == 1:
            return values[0][1]
        else:
            return map(lambda x: x[1], values)
    else:
        return None

def handle_itip_part(part, recipient):
    method = part.get_param("method")

    # Decode the data and parse it.

    f = StringIO(part.get_payload(decode=True))

    try:
        doctype, attrs, elements = parse(f, encoding=part.get_content_charset())
    except (ParseError, ValueError):
        sys.exit(EX_DATAERR)

    # Only handle calendar information.

    if doctype == "VCALENDAR":
        itip = get_itip_elements(elements)

        # Require consistency between declared and employed methods.

        if get_value(itip, "METHOD") == method:

            # Look for different kinds of sections.

            for name, cls in handlers:
                for details in get_value(itip, name, False) or []:

                    # Dispatch to a handler and obtain any response.

                    handler = cls(details)
                    part = methods[method](handler)()

                    if part:
                        print >>open("/tmp/imip.txt", "a"), part.as_string()

class Handler:

    "General handler support."

    def __init__(self, details):
        self.details = details
        self.uid = get_value(details, "UID")
        self.sequence = get_value(details, "SEQUENCE")
        self.store = imip_store.FileStore()

    def get_attr_value(self, name, single=True):
        return get_attr_value(self.details, name, single)

    def get_value(self, name, single=True):
        return get_value(self.details, name, single)

class Event(Handler):

    "An event handler."

    def add(self):
        pass

    def cancel(self):
        pass

    def counter(self):
        pass

    def declinecounter(self):
        pass

    def publish(self):
        pass

    def refresh(self):
        pass

    def reply(self):

        "Since this handler does not send requests, it will not handle replies."

        pass

    def request(self):
        pass

class Freebusy(Handler):

    "A free/busy handler."

    def publish(self):
        pass

    def reply(self):

        "Since this handler does not send requests, it will not handle replies."

        pass

    def request(self):

        """
        Respond to a request by sending a reply containing free/busy information
        for each indicated attendee.
        """

        attendees = self.get_attr_value("ATTENDEE", False)
        organiser = self.get_attr_value("ORGANIZER")

        if not attendees and not organiser:
            return

        organiser_attr, organiser = organiser

        # Get the details for the attendee.

        out = StringIO()
        try:
            w = iterwrite(out, encoding="utf-8")

            calendar = []
            cwrite = calendar.append
            cwrite(("METHOD", {}, "REPLY"))
            cwrite(("VERSION", {}, "2.0"))

            for attendee_attr, attendee in attendees:
                freebusy = self.store.get_freebusy(attendee)
                if freebusy:
                    record = []
                    rwrite = record.append
                    rwrite(("ORGANIZER", organiser_attr, organiser))
                    rwrite(("ATTENDEE", attendee_attr, attendee))
                    rwrite(("UID", {}, self.uid))

                    for start, end in freebusy:
                        rwrite(("FREEBUSY", {}, [start, end]))

                    cwrite(("VFREEBUSY", {}, record))

            # Send a reply with the information.

            w.write("VCALENDAR", {}, calendar)

            return MIMEText(out.getvalue(), "calendar", "utf-8")

        finally:
            out.close()

class Journal(Handler):

    "A journal entry handler."

    def add(self):
        pass

    def cancel(self):
        pass

    def publish(self):
        pass

class Todo(Handler):

    "A to-do item handler."

    def add(self):
        pass

    def cancel(self):
        pass

    def counter(self):
        pass

    def declinecounter(self):
        pass

    def publish(self):
        pass

    def refresh(self):
        pass

    def reply(self):

        "Since this handler does not send requests, it will not handle replies."

        pass

    def request(self):
        pass

# Handler registry.

handlers = [
    ("VFREEBUSY",   Freebusy),
    ("VEVENT",      Event),
    ("VTODO",       Todo),
    ("VJOURNAL",    Journal),
    ]

methods = {
    "ADD"            : lambda handler: handler.add,
    "CANCEL"         : lambda handler: handler.cancel,
    "COUNTER"        : lambda handler: handler.counter,
    "DECLINECOUNTER" : lambda handler: handler.declinecounter,
    "PUBLISH"        : lambda handler: handler.publish,
    "REFRESH"        : lambda handler: handler.refresh,
    "REPLY"          : lambda handler: handler.reply,
    "REQUEST"        : lambda handler: handler.request,
    }

if __name__ == "__main__":
    process(sys.stdin, sys.argv[1:])

# vim: tabstop=4 expandtab shiftwidth=4
